// Organism.cc

#include "Organism.h"
#include "Simulator.h"
#include "EnvVector.h"
#include "NRCRandom.h"
#include "OrgID.h"
#include <cmath>
#include <iterator>
#include <algorithm>
#include <fstream>


Organism::Organism( int timeOfBirth /* = 0 */, bool bRandomCreate /* = true */, Simulator * pSim /* = NULL */ )
:
	m_ReqStartPos( 0 ),
	m_Phenotype( Simulator::LOCAL_ENVIRONMENT_VECTOR_SIZE, 0.0f ),
	m_Fitness( 0.0f ),
	m_ProbDeath( Simulator::DEATH_PROB_COEFF_X ),
	m_ProbReproduction( 0.0f ),
	m_Specificity( 1 ),
	m_SpecificityFraction( 1.0f / (float)Simulator::LOCAL_ENVIRONMENT_VECTOR_SIZE ),
	m_TimeOfBirth( timeOfBirth ),
	m_NumOffspring( 0 ),
	m_bGenomeRecorded( false )
{
	if (bRandomCreate)
		randomCreate( pSim );
}


Organism::Organism( Organism * pParent, int timeOfBirth, Simulator * pSim, bool bMutate = true )
:
	m_ReqStartPos( pParent->m_ReqStartPos ),
	m_Genome( pParent->m_Genome ),
	m_EnvDisturbance( pParent->m_EnvDisturbance ),
	m_Phenotype( Simulator::LOCAL_ENVIRONMENT_VECTOR_SIZE, 0.0f ),
	m_Fitness( 0.0f ),
	m_ProbDeath( Simulator::DEATH_PROB_COEFF_X ),
	m_ProbReproduction( 0.0f ),
	m_Specificity( pParent->m_Specificity ),
	m_SpecificityFraction( pParent->m_SpecificityFraction ),
	m_ParentID( pParent->m_ID ),
	m_TimeOfBirth( timeOfBirth ),
	m_NumOffspring( 0 ),
	m_bGenomeRecorded( false )
{
	pParent->recordNewOffspring();
	
	int numMutations = 0;

	if (bMutate)
	{
		// add genes
		if ( Simulator::m_RNG.rand() < Simulator::MUTATION_ADD_PROB )
		{
			m_Genome.push_back( Gene(true) );
			++numMutations;
		}

		// delete genes
		if ( Simulator::m_RNG.rand() < Simulator::MUTATION_DELETE_PROB )
		{
			int numGenes = m_Genome.size();
			if (numGenes > 1)
			{
				int geneNum = Simulator::m_RNG.randInt() % numGenes;
				vector<Gene>::iterator gen_it = m_Genome.begin();
				advance( gen_it, geneNum );
				m_Genome.erase( gen_it );
				++numMutations;
			}
		}
		
		// mutate Genome
		int numGenes = m_Genome.size();
		float mutationRate = Simulator::MUTATION_CHANGE_PROB_PER_GENE;
		if (Simulator::MUTATION_PROPORTIONAL_TO_LENGTH)
			mutationRate /= (float)numGenes;
		int numGeneMutations = (int) ( bnldev( mutationRate, numGenes, &Simulator::NRC_RNG_STATE ) + Simulator::FLOATING_POINT_ACCURACY );
		vector<int> geneIdx( numGenes );
		for (int i = 0; i < numGenes; ++i)
			geneIdx[i] = i;
		Simulator::RNGFunctionObject rng;		
		random_shuffle( geneIdx.begin(), geneIdx.end(), rng );
		for (int i = 0; i < numGeneMutations; ++i)
			m_Genome[geneIdx[i]].mutate();
		numMutations += numGeneMutations;
		
		// mutate EnvDisturbance
		if ( Simulator::m_RNG.rand() < Simulator::MUTATION_CHANGE_PROB_PER_GENE )
		{
			m_EnvDisturbance.mutate( true );
			++numMutations;
		}
		
		// mutate Requirements
		if ( Simulator::ORGANISM_REQUIREMENTS_EVOLVABLE )
		{
			if ( Simulator::m_RNG.rand() < Simulator::MUTATION_CHANGE_PROB_REQ )
			{
				int delta = 1 - (Simulator::m_RNG.randInt() % 3);
				m_ReqStartPos += delta;
				
				if (delta != 0)
				{
					++numMutations;
					if (m_ReqStartPos < 0)
						m_ReqStartPos += Simulator::LOCAL_ENVIRONMENT_VECTOR_SIZE;
					else if (m_ReqStartPos >= Simulator::LOCAL_ENVIRONMENT_VECTOR_SIZE)
						m_ReqStartPos -= Simulator::LOCAL_ENVIRONMENT_VECTOR_SIZE;
				}
				
				delta = 1 - (Simulator::m_RNG.randInt() % 3);
				m_Specificity += delta;
				
				if (delta != 0)
				{
					++numMutations;
					if (m_Specificity <= 0)
						m_Specificity = 1;
					else if (m_Specificity > Simulator::LOCAL_ENVIRONMENT_VECTOR_SIZE)
						m_Specificity = Simulator::LOCAL_ENVIRONMENT_VECTOR_SIZE;
				}
			}
		}

		// finally, update any other state data that might have changed
		updateSpecificity();
		updatePhenotype();
	}
		
	// last but not least, decide on a name
	if (numMutations > 0)
	{
		pSim->getNewOrgID( m_Specificity, m_Genome.size(), &m_ID );
		if (pSim->inPhylogenyRecordWindow())
		{
			ofstream phylogeny( "phylogeny.dat", ios::app );
			phylogeny << m_ID << "," << m_ParentID << endl;
			phylogeny.close();
		}
		m_bGenomeRecorded = false;
	}
	else
	{
		m_ID = m_ParentID;
		m_bGenomeRecorded = pParent->m_bGenomeRecorded;
	}
}


void Organism::randomCreate( Simulator * pSim /* = NULL */ )
{
	m_ReqStartPos = Simulator::m_RNG.randInt() % Simulator::LOCAL_ENVIRONMENT_VECTOR_SIZE;
	m_Specificity = (int)((Simulator::ORGANISM_SPECIFICITY_FRACTION_INIT * (float)Simulator::LOCAL_ENVIRONMENT_VECTOR_SIZE) + Simulator::FLOATING_POINT_ACCURACY);
	
	m_SpecificityFraction = (float)m_Specificity / (float)Simulator::LOCAL_ENVIRONMENT_VECTOR_SIZE;

	m_Genome.clear();
	
	int numGenes = Simulator::NEW_GENOME_LENGTH_MIN;
	if (Simulator::NEW_GENOME_LENGTH_MAX > Simulator::NEW_GENOME_LENGTH_MIN)
		numGenes += (Simulator::m_RNG.randInt() % (1 + Simulator::NEW_GENOME_LENGTH_MAX - Simulator::NEW_GENOME_LENGTH_MIN) );
	
	for (int i = 0; i < numGenes; ++i)
		m_Genome.push_back( Gene(true) );
	
	m_EnvDisturbance.randomCreate( true );
	
	updatePhenotype();
	
	if (pSim != NULL)
	{
		pSim->getNewOrgID( m_Specificity, m_Genome.size(), &m_ID );
		if (pSim->inPhylogenyRecordWindow())
		{
			ofstream phylogeny( "phylogeny.dat", ios::app );
			phylogeny << m_ID << "," << m_ParentID << endl;
			phylogeny.close();
		}
		m_bGenomeRecorded = false;
	}
}


void Organism::updateSpecificity()
{
	m_SpecificityFraction = (float)m_Specificity / (float)Simulator::LOCAL_ENVIRONMENT_VECTOR_SIZE;	
}


void Organism::calculateFitness( EnvVector & env )
{
	m_Fitness = 1.0f - ( hammingDistance( env ) / (float)m_Specificity ); // fitness is between 0 and 1

	switch ( Simulator::REPRODUCTION_FUNCTION_TYPE )
	{
		case Simulator::rfLinear:
		{
			m_ProbReproduction = Simulator::REPRODUCTION_FUNCTION_COEFF_X * m_Fitness;
			break;
		}
		case Simulator::rfPower:
		{
			// prob is a power function of fitness, REPRODUCTION_FUNCTION_COEFF_K should be > 0
			m_ProbReproduction = Simulator::REPRODUCTION_FUNCTION_COEFF_X * pow( m_Fitness, Simulator::REPRODUCTION_FUNCTION_COEFF_K );
			break;
		}
		case Simulator::rfSigmoid:
		{
			// prob is a sigmoid function of fitness, with prob = 0.5 when fitness = REPRODUCTION_FUNCTION_COEFF_C
			m_ProbReproduction = Simulator::REPRODUCTION_FUNCTION_COEFF_X / (1.0f + exp(-Simulator::REPRODUCTION_FUNCTION_COEFF_K * (m_Fitness - Simulator::REPRODUCTION_FUNCTION_COEFF_C)));
			break;
		}
		default:
		{
			cerr << "In Organism::calculateFitness. Encountered unknown reproduction function " << Simulator::REPRODUCTION_FUNCTION_TYPE << "! Aborting." << endl;
			exit( EXIT_FAILURE );
		}		
	}

	switch ( Simulator::DEATH_POLICY )
	{
		case Simulator::dpConstant:
		{
			m_ProbDeath = Simulator::DEATH_PROB_COEFF_X;
			break;
		}
		case Simulator::dpOneMinusFitness:
		{
			m_ProbDeath = Simulator::DEATH_PROB_COEFF_X * (1.0f - m_Fitness);
			break;
		}
		case Simulator::dpOneMinusFitnessTimesOneMinusSpec:
		{
			m_ProbDeath = Simulator::DEATH_PROB_COEFF_X * (1.0f - m_Fitness) * (1.0f - m_SpecificityFraction);
			break;
		}
		case Simulator::dpOneMinusProbRep:
		{
			m_ProbDeath = Simulator::DEATH_PROB_COEFF_X * (1.0f - m_ProbReproduction);
			break;
		}
		case Simulator::dpOneMinusProbRepTimesOneMinusSpec:
		{
			m_ProbDeath = Simulator::DEATH_PROB_COEFF_X * (1.0f - m_ProbReproduction) * (1.0f - m_SpecificityFraction);
			break;
		}
		default:
		{
			cerr << "In Organism::calculateFitness. Encountered unknown death policy " << Simulator::DEATH_POLICY << "! Aborting." << endl;
			exit( EXIT_FAILURE );
		}
	}

	/*
	fit		spec	pd (1/fs)	pd (1-fs)	pd (1-f)(1-s)
	0.1		0.1 	100			0.99		0.81 
	0.1		0.5		20			0.95		0.45
	0.1		0.9		11			0.91		0.09
	0.5		0.1		20			0.95		0.45
	0.5		0.5		4			0.75		0.25
	0.5		0.9		2			0.55		0.05
	0.9		0.1		11			0.91		0.09
	0.9		0.5		2			0.55		0.05
	0.9		0.9		1			0.19		0.01
	*/
}


float Organism::hammingDistance( EnvVector & env )
{
	//assert( m_Phenotype.length() == Simulator::LOCAL_ENVIRONMENT_VECTOR_SIZE );
	//assert( env.length() == Simulator::LOCAL_ENVIRONMENT_VECTOR_SIZE );
	//assert( (int)m_Requirements.size() == Simulator::LOCAL_ENVIRONMENT_VECTOR_SIZE );

	float hamming = 0.0f;

	for (int i = m_ReqStartPos; i < (m_ReqStartPos + m_Specificity); ++i)
	{
		int idx = i;
		
		if (idx < 0)
			idx += Simulator::LOCAL_ENVIRONMENT_VECTOR_SIZE;
		else if (idx >= Simulator::LOCAL_ENVIRONMENT_VECTOR_SIZE)
			idx -= Simulator::LOCAL_ENVIRONMENT_VECTOR_SIZE;

		hamming += fabs(m_Phenotype.getValues()[idx] - env.getValues()[idx]);
	}

	return hamming;
}


void Organism::updatePhenotype()
{
	vector<float>::iterator phen_it  = m_Phenotype.getValues().begin();
	vector<float>::iterator phen_end = m_Phenotype.getValues().end();
	while (phen_it != phen_end)
	{
		*phen_it = Simulator::LOCAL_ENVIRONMENT_VECTOR_INIT_VAL;
		++phen_it;
	}
	
	vector<Gene>::iterator gen_it  = m_Genome.begin();
	vector<Gene>::iterator gen_end = m_Genome.end();
	while (gen_it != gen_end)
	{
		m_Phenotype.modify(&(*gen_it));
		++gen_it;
	}
}


void Organism::recordNewOffspring()
{
	++m_NumOffspring;
	if (Simulator::RECORD_GENOME_BANK_FILE && (m_NumOffspring == Simulator::GENOME_BANK_RECORD_THRESHOLD))
	{
		if (!m_bGenomeRecorded)
		{
			ofstream genomebank( "genomebank.dat", ios::app );
			genomebank << *this << endl;
			genomebank.close();
			m_bGenomeRecorded = true;
		}
	}
}

ostream& operator << (ostream& os, Organism& id)
{
	os	<< id.m_ID << " " << id.m_ParentID << " " << id.m_TimeOfBirth << " " << id.m_NumOffspring << " "
		<< id.m_Specificity << " " << id.m_SpecificityFraction << " "
		<< id.m_ReqStartPos << " " << id.m_Fitness << " "
		<< id.m_ProbDeath << " " << id.m_ProbReproduction << " ";
	
	os << id.m_Genome.size() << " ";
	vector<Gene>::iterator gen_it  = id.m_Genome.begin();
	vector<Gene>::iterator gen_end = id.m_Genome.end();
	while (gen_it != gen_end)
	{
		os << *gen_it << " ";
		++gen_it;
	}
	
	os << id.m_EnvDisturbance << " " << id.m_Phenotype;
	
	return os;
}
