// Patch.cc

#include "Patch.h"
#include "Organism.h"
#include "Simulator.h"
#include "Gene.h"
#include "EnvVector.h"
#include <iostream>
#include <fstream> // only required for Patch::test
#include <sstream> // only required for Patch::test


Patch::Patch()
:
	m_LocalAbioticEnv( Simulator::LOCAL_ENVIRONMENT_VECTOR_SIZE, Simulator::LOCAL_ENVIRONMENT_VECTOR_INIT_VAL ),
	m_pSim(NULL),
	m_X(0),
	m_Y(0)
{
	if ( Simulator::ABIOTIC_ENV_TYPE == Simulator::aetComplex )
	{
		// order of gene constructor params: fn, prt, spr, rep, op
		Gene gene1( 0, 4, 2, 3, 5 );
		Gene gene2( 3, 6, 1, 0, 3 );
		Gene gene3( 4, 2, 0, 1, 1 );
		Gene gene4( 6, 3, 4, 1, 8 );
		Gene gene5( 10, 7, 2, 0, 12 );
		m_LocalAbioticEnv.modify( &gene1 );
		m_LocalAbioticEnv.modify( &gene2 );
		m_LocalAbioticEnv.modify( &gene3 );
		m_LocalAbioticEnv.modify( &gene4 );
		m_LocalAbioticEnv.modify( &gene5 );
	}
	
	m_CurrentModifiedEnv = m_LocalAbioticEnv;
}


Patch::~Patch()
{
	vector<Organism *>::iterator org_it  = m_Organisms.begin();
	vector<Organism *>::iterator org_end = m_Organisms.end();
	while (org_it != org_end)
	{
		delete (*org_it);
		++org_it;
	}
}


void Patch::addOrganism( Organism * pOrg )
{
	m_Organisms.push_back( pOrg );
	m_pSim->recordOrgBirth( pOrg->getID() );
}


void Patch::updateEnvironment()
{
	int divisor = 1;

	m_CurrentModifiedEnv = m_LocalAbioticEnv;
	
	for (int ix = - Simulator::NEIGHBOURHOOD_RADIUS; ix <= Simulator::NEIGHBOURHOOD_RADIUS; ++ix)
	{
		for (int iy = - Simulator::NEIGHBOURHOOD_RADIUS; iy <= Simulator::NEIGHBOURHOOD_RADIUS; ++iy)
		{
			Patch & rNeighbour = getNeighbour(ix,iy);
			divisor += rNeighbour.addEnvDisturbances( &m_CurrentModifiedEnv );
		}
	}
	
	m_CurrentModifiedEnv.divide( (float)divisor );
}


inline Patch & Patch::getNeighbour(int ix, int iy)
{
	return m_pSim->getPatchFW(m_X+ix, m_Y+iy);
}


int Patch::addEnvDisturbances( EnvVector * pEnv )
{
	vector<Organism*>::iterator it_org = m_Organisms.begin();
	vector<Organism*>::iterator it_end = m_Organisms.end();
	while (it_org != it_end)
	{
		pEnv->add(&((*it_org)->getEnvDisturbance()));
		++it_org;
	}
	
	return (int)(m_Organisms.size());
}


void Patch::updateOrganisms()
{
	vector<Organism*>::iterator it_org = m_Organisms.begin();
	vector<Organism*>::iterator it_end = m_Organisms.end();
	while (it_org != it_end)
	{
		(*it_org)->calculateFitness( m_CurrentModifiedEnv );
		++it_org;
	}
}


void Patch::killOrganismsStochastic()
{
	vector<Organism*>::iterator it_org = m_Organisms.begin();

	while (it_org != m_Organisms.end())
	{
		Organism * pOrg = *it_org;
		if ( Simulator::m_RNG.rand() < pOrg->getProbDeath() )
		{
			m_pSim->recordOrgDeath( pOrg );
			delete pOrg;
			it_org = m_Organisms.erase( it_org );
		}
		else
		{
			++it_org;
		}
	}
}


void Patch::killOrganism()
{
	float maxProb = 0.0f;
	int idx = 0;
	int numOrgs = m_Organisms.size();
	
	for (int i = 0; i < numOrgs; ++i)
	{
		if (m_Organisms[i]->getProbDeath() > maxProb)
		{
			maxProb = m_Organisms[i]->getProbDeath();
			idx = i;
		}
	}
	
	vector<Organism *>::iterator org_it = m_Organisms.begin();
	advance( org_it, idx );
	m_pSim->recordOrgDeath( *org_it );
	delete *org_it;
	m_Organisms.erase( org_it );
}


void Patch::killOrganismRouletteWheel()
{
	int numOrgs = m_Organisms.size();	
	float wheelSize = 0.0f;
	for (int i = 0; i < numOrgs; ++i)
		wheelSize += m_Organisms[i]->getProbDeath();
	
	float shot = Simulator::m_RNG.randExc() * wheelSize;
	float cumulativeProb = 0.0f;
	int idx = -1;
	
	for (int i = 0; ((i < numOrgs) && (idx == -1)); ++i)
	{
		cumulativeProb += m_Organisms[i]->getProbDeath();
		if (cumulativeProb >= shot)
			idx = i;
	}

	assert( idx >= 0 );
	
	vector<Organism *>::iterator org_it = m_Organisms.begin();
	advance( org_it, idx );
	m_pSim->recordOrgDeath( *org_it );
	delete *org_it;
	m_Organisms.erase( org_it );
}


void Patch::reproduceOrganisms( int iteration )
{
	int numOrgs = m_Organisms.size();

	for (int i = 0; i < numOrgs; ++i)
	{
		Organism * pOrg = m_Organisms[i];
		if ( Simulator::m_RNG.rand() < pOrg->getProbReproduction() )
		{
			Organism * pNewOrg = new Organism( pOrg, iteration, m_pSim, true );
			int x = m_X - Simulator::DISPERSAL_RADIUS + (Simulator::m_RNG.randInt() % (2 * Simulator::DISPERSAL_RADIUS + 1));
			int y = m_Y - Simulator::DISPERSAL_RADIUS + (Simulator::m_RNG.randInt() % (2 * Simulator::DISPERSAL_RADIUS + 1));
			m_pSim->getPatchFW(x,y).addOrganism( pNewOrg );
		}
	}
}


void Patch::test()
{
	// test gene action on environment vectors
	
	for (int i=0; i<20; ++i)
	{
		Gene gene(true);
		m_CurrentModifiedEnv = m_LocalAbioticEnv;
		m_CurrentModifiedEnv.modify( &gene );

		stringstream name;
		name << "geneF" << gene.m_Function << "P" << gene.m_Partition << "S"
			<< gene.m_Spread << "R" << gene.m_Repeat << "O" << gene.m_Operator;
		ofstream file(name.str().c_str());
		vector<float> & vec = m_CurrentModifiedEnv.getValues();
		for (unsigned int j=0; j<vec.size(); ++j)
			file << vec[j] << " ";
		file << endl;
		file.close();
	}
	
}


ostream& operator << (ostream& os, const Patch& id)
{
	os << id.m_X << "," << id.m_Y << " " << id.m_Organisms.size() << " ";
	vector<Organism*>::const_iterator org_it  = id.m_Organisms.begin();
	vector<Organism*>::const_iterator org_end = id.m_Organisms.end();
	while (org_it != org_end)
	{
		os << *(*org_it) << " ";
		++org_it;
	}
	
	os << id.m_LocalAbioticEnv << " " << id.m_CurrentModifiedEnv;
	
	return os;
}
