// Simulator.cc

#include "Simulator.h"
#include "Organism.h"
#include "MersenneTwister.h"
#include "NRCRandom.h"
#include "Gene.h"
#include <cassert>
#include <utility>
#include <algorithm>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <cstdlib>
#include <ctime>


MTRand Simulator::m_RNG;
long Simulator::NRC_RNG_STATE = -1;

// statics associated with parameter registry
int Simulator::numIntParams = 0;
int Simulator::numUIntParams = 0;
int Simulator::numULongParams = 0;
int Simulator::numFloatParams = 0;
int Simulator::numBoolParams = 0;
int Simulator::numStringParams = 0;
int *Simulator::IntParams[_MaxIntParams];
char Simulator::IntParamNames[_MaxIntParams][_MaxParamNameLength];
unsigned int *Simulator::UIntParams[_MaxUIntParams];
char Simulator::UIntParamNames[_MaxUIntParams][_MaxParamNameLength];
unsigned long *Simulator::ULongParams[_MaxULongParams];
char Simulator::ULongParamNames[_MaxULongParams][_MaxParamNameLength];
float *Simulator::FloatParams[_MaxFloatParams];
char Simulator::FloatParamNames[_MaxFloatParams][_MaxParamNameLength];
bool *Simulator::BoolParams[_MaxBoolParams];
char Simulator::BoolParamNames[_MaxBoolParams][_MaxParamNameLength];
char *Simulator::StringParams[_MaxStringParams];
char Simulator::StringParamNames[_MaxStringParams][_MaxParamNameLength];

// initialise parameters
unsigned long Simulator::RNG_SEED = 0;
int Simulator::NUM_ITERATIONS = 10000;
int Simulator::ENV_SIZE_X = 100;
int Simulator::ENV_SIZE_Y = 100;
int Simulator::DISPERSAL_RADIUS = 2;
int Simulator::NEIGHBOURHOOD_RADIUS = 1;
Simulator::eReproductionFunction Simulator::REPRODUCTION_FUNCTION_TYPE = Simulator::rfLinear;
float Simulator::REPRODUCTION_FUNCTION_COEFF_K = 10.0f;
float Simulator::REPRODUCTION_FUNCTION_COEFF_C = 0.3f;
float Simulator::REPRODUCTION_FUNCTION_COEFF_X = 1.0f;
Simulator::eDeathPolicy Simulator::DEATH_POLICY = Simulator::dpOneMinusProbRep;
float Simulator::DEATH_PROB_COEFF_X = 0.05f;
float Simulator::LOCAL_POP_DENSITY_MAX = 5.0f;
float Simulator::LOCAL_POP_DENSITY_INIT = 0.1f;
float Simulator::MUTATION_ADD_PROB = 0.05f;
float Simulator::MUTATION_DELETE_PROB = 0.05f;
float Simulator::MUTATION_CHANGE_PROB_PER_GENE = 0.05f;
float Simulator::MUTATION_CHANGE_PROB_REQ = 0.05f;
bool Simulator::MUTATION_PROPORTIONAL_TO_LENGTH = true;
bool Simulator::ORGANISM_REQUIREMENTS_EVOLVABLE = true; 
bool Simulator::ENVIRONMENT_DISTURBANCE_ENABLED = true;
bool Simulator::ENVIRONMENT_DISTURBANCE_RESTRICTED = false;
int Simulator::ENVIRONMENT_DISTURBANCE_MAX_SPREAD = 0;
float Simulator::ORGANISM_SPECIFICITY_FRACTION_INIT = 0.20f;
int Simulator::LOCAL_ENVIRONMENT_VECTOR_SIZE = 100;
float Simulator::LOCAL_ENVIRONMENT_VECTOR_INIT_VAL = 0.5f;
Simulator::eAbioticEnvType Simulator::ABIOTIC_ENV_TYPE = Simulator::aetSimple;
int Simulator::NEW_GENOME_LENGTH_MIN = 1;
int Simulator::NEW_GENOME_LENGTH_MAX = 1;

bool Simulator::RECORD_PHYLOGENY_FILE_FULL = false;
bool Simulator::RECORD_PHYLOGENY_FILE_PARTIAL = true;
bool Simulator::RECORD_GENOME_BANK_FILE = true;
bool Simulator::RECORD_GENOME_SIZE_FILE = true;
bool Simulator::RECORD_GENOME_SPECIFICITY_FILE = true;
bool Simulator::RECORD_SPECIES_COUNTS_FILE = true;
bool Simulator::RECORD_SIMULATION_STATE = true;
int Simulator::DATA_RECORD_PERIOD = 5;
int Simulator::SIMULATION_STATE_RECORD_PERIOD = 1000;
int Simulator::PHYLOGENY_PARTIAL_RECORD_PERIOD = 1000;
int Simulator::PHYLOGENY_PARTIAL_RECORD_WINDOW = 100;
int Simulator::GENOME_SIZE_FILE_MAX_SIZE = 50;
int Simulator::GENOME_BANK_RECORD_THRESHOLD = 10;
int Simulator::SPECIES_NUMBER_SIG_THRESHOLD = 25;
int Simulator::SPECIES_COUNTS_FILE_ZIP_PERIOD = 500;
int Simulator::NEIGHBOURHOOD_SIZE = (NEIGHBOURHOOD_RADIUS * 2 + 1) * (NEIGHBOURHOOD_RADIUS * 2 + 1);



Simulator::Simulator()
:
	m_Environment( ENV_SIZE_X * ENV_SIZE_Y ),
	m_CurrentIteration( 0 ),
	m_LatestSpeciesCountsIdx( 0 )
{
	for (int x = 0 ; x < ENV_SIZE_X; ++x)
		for (int y = 0; y < ENV_SIZE_Y; ++y)
			getPatch(x,y).initialise(this,x,y);

	initialise();
}


Simulator::~Simulator()
{
}


void Simulator::initialise()
{
	if (RNG_SEED == 0)
		RNG_SEED = time( NULL );
	
	m_RNG.seed( RNG_SEED );
	NRC_RNG_STATE = -(long)RNG_SEED;
	ran1( &NRC_RNG_STATE );
	
	
	ofstream info( "runinfo.txt", ios::app );
	if (! info.good())
	{
		cerr << "Unable to open file runinfo.txt for writing!" << endl;
		exit( EXIT_FAILURE );
	}
	recordParameterValues( info );
	info.close();
	
	int numOrgs = (int) (((float)LOCAL_POP_DENSITY_INIT * (float)ENV_SIZE_X * (float)ENV_SIZE_Y) + FLOATING_POINT_ACCURACY);
	
	assert( LOCAL_POP_DENSITY_INIT < LOCAL_POP_DENSITY_MAX );
	
	for (int i = 0; i < numOrgs; ++i)
	{
		Organism * pOrg = new Organism( 0, true, this );

		int x = m_RNG.randInt() % ENV_SIZE_X;
		int y = m_RNG.randInt() % ENV_SIZE_Y;

		while (getLocalDensity(x,y,true) > LOCAL_POP_DENSITY_MAX)
		{
			x = m_RNG.randInt() % ENV_SIZE_X;
			y = m_RNG.randInt() % ENV_SIZE_Y;
		}

		getPatch(x,y).addOrganism(pOrg);
	}
}


float Simulator::getLocalDensity( int x, int y, bool plusOne /* = false */ )
{
	int numOrgs = plusOne ? 1 : 0;

	for (int ix = x - NEIGHBOURHOOD_RADIUS; ix <= x + NEIGHBOURHOOD_RADIUS; ++ix)
		for (int iy = y - NEIGHBOURHOOD_RADIUS; iy <= y + NEIGHBOURHOOD_RADIUS; ++iy)
			numOrgs += getPatchFW(ix,iy).getNumOrganisms();
		
	return (((float)numOrgs) / (float)NEIGHBOURHOOD_SIZE);
}


void Simulator::run( int maxIterations )
{
	int x, y;
	
	recordSummaryData( 0 );
	
	for (m_CurrentIteration = 0; m_CurrentIteration < maxIterations; ++m_CurrentIteration)
	{
		if ( ENVIRONMENT_DISTURBANCE_ENABLED )
		{
			for (x = 0; x < ENV_SIZE_X; ++x)
				for (y = 0; y < ENV_SIZE_Y; ++y)
					getPatch(x,y).updateEnvironment();			// update environment vector according to organisms present
		}
			
		for (x = 0; x < ENV_SIZE_X; ++x)
		{
			for (y = 0; y < ENV_SIZE_Y; ++y)
			{
				Patch & patch = getPatch(x,y);
				patch.updateOrganisms();						// recalculate fitness etc.
				patch.killOrganismsStochastic();				// stochastically kill some organisms
				patch.reproduceOrganisms( m_CurrentIteration );	// stochastically reproduce organisms
			}
		}

		checkOvercrowding();									// kill more organisms in overcrowded areas
		
		if  ((maxIterations-1==m_CurrentIteration) ||
			(1==DATA_RECORD_PERIOD) ||
			((m_CurrentIteration > 0) && (0==m_CurrentIteration % DATA_RECORD_PERIOD)))
			recordSummaryData( m_CurrentIteration+1 );

		if  ((RECORD_SIMULATION_STATE) &&
				((maxIterations-1==m_CurrentIteration) ||
				 (1==SIMULATION_STATE_RECORD_PERIOD) ||
				 ((m_CurrentIteration > 0) && (0==m_CurrentIteration % SIMULATION_STATE_RECORD_PERIOD))))
			saveState( m_CurrentIteration+1 );
	}
}


void Simulator::checkOvercrowding()
{
	int x, y, ix, iy, k;
	RNGFunctionObject rng;

	for (x = 0; x < ENV_SIZE_X; ++x)
	{
		for (y = 0; y < ENV_SIZE_Y; ++y)
		{
			float localDensity = getLocalDensity(x,y);
			
			if (localDensity > LOCAL_POP_DENSITY_MAX)
			{
				vector< pair<int,int> > org_pos;
				for (ix = x - NEIGHBOURHOOD_RADIUS; ix <= x + NEIGHBOURHOOD_RADIUS; ++ix)
				{
					for (iy = y - NEIGHBOURHOOD_RADIUS; iy <= y + NEIGHBOURHOOD_RADIUS; ++iy)
					{
						int numOrgs = getPatchFW(ix,iy).getNumOrganisms();
						pair<int, int> pos(fwx(ix),fwy(iy));
						org_pos.insert( org_pos.end(), numOrgs, pos ); // insert numOrgs copies of pos
					}
				}

				random_shuffle( org_pos.begin(), org_pos.end(), rng );
				
				int numToKill = (int) ((localDensity - LOCAL_POP_DENSITY_MAX + FLOATING_POINT_ACCURACY) * (float)NEIGHBOURHOOD_SIZE);
				for (k = 0; k < numToKill; ++k)
				{
					ix = org_pos[k].first;
					iy = org_pos[k].second;
					getPatch(ix,iy).killOrganismRouletteWheel();
				}						
			}
		}
	}	
}


void Simulator::recordSummaryData( int iteration )
{
	int totalNumOrgs = 0;
	int totalNumSigOrgs = 0;
	int numOrgs = 0;
	//int numSigOrgsWithDeaths = 0;

	int diversity = 0;
	int sigDiversity = 0;	
	
	int orgSize = 0;
	int minSize = 999;
	int maxSize = 0;
	unsigned int meanSize = 0;
	int minSigSize = 999;
	int maxSigSize = 0;
	unsigned int meanSigSize = 0;	
	
	int spec = 0;
	int minSpecificity = LOCAL_ENVIRONMENT_VECTOR_SIZE;
	int maxSpecificity = 0;
	unsigned int meanSpecificity = 0;
	int minSigSpecificity = LOCAL_ENVIRONMENT_VECTOR_SIZE;
	int maxSigSpecificity = 0;
	unsigned int meanSigSpecificity = 0;
	
	int age = 0;
	float minAge = 999;
	float maxAge = 0;
	float meanAge = 0;
	float minSigAge = 999;
	float maxSigAge = 0;
	float meanSigAge = 0;
	
	int numOffspring = 0;
	float minNumOffspring = 999;
	float maxNumOffspring = 0;
	float meanNumOffspring = 0;
	float minSigNumOffspring = 999;
	float maxSigNumOffspring = 0;
	float meanSigNumOffspring = 0;	

	float probRep = 0.0f;
	float minProbRep = 1.1f;
	float maxProbRep = 0.0f;
	float meanProbRep = 0.0f;
	float minSigProbRep = 1.1f;
	float maxSigProbRep = 0.0f;
	float meanSigProbRep = 0.0f;	

	vector<int> numgenecounts( GENOME_SIZE_FILE_MAX_SIZE, 0 );
	vector<int> speccounts( LOCAL_ENVIRONMENT_VECTOR_SIZE, 0 );
	
	for (int x = 0; x < ENV_SIZE_X; ++x)
	{
		for (int y = 0; y < ENV_SIZE_Y; ++y)
		{
			vector<Organism *> & orgs = getPatch(x,y).getOrganisms();
			numOrgs = orgs.size();
			for (int i=0; i < numOrgs; ++i)
			{
				++totalNumOrgs;
				
				// calculate size stats
				orgSize = orgs[i]->getNumGenes();
				
				meanSize += orgSize;
				
				if (orgSize > maxSize)
					maxSize = orgSize;
				
				if (orgSize < minSize)
					minSize = orgSize;
				
				if (orgSize > GENOME_SIZE_FILE_MAX_SIZE)
					orgSize = GENOME_SIZE_FILE_MAX_SIZE;
				numgenecounts[orgSize-1] = numgenecounts[orgSize-1] + 1;
				
				// calculate specificity stats
				spec = orgs[i]->getSpecificity();
				
				meanSpecificity += spec;
				
				if (spec > maxSpecificity)
					maxSpecificity = spec;
				
				if (spec < minSpecificity)
					minSpecificity = spec;
				
				speccounts[spec-1] = speccounts[spec-1] + 1;
				
				// calculate age stats
				age = 1 + m_CurrentIteration - orgs[i]->getTimeOfBirth();
				
				meanAge += age;
				
				if (age > maxAge)
					maxAge = age;
				
				if (age < minAge)
					minAge = age;
					
				// calculate num offspring stats
				numOffspring = orgs[i]->getNumOffspring();
				
				meanNumOffspring += numOffspring;
			
				if (numOffspring > maxNumOffspring)
					maxNumOffspring = numOffspring;
			
				if (numOffspring < minNumOffspring)
					minNumOffspring = numOffspring;				

				// calculate probability of reproduction stats
				probRep = orgs[i]->getProbReproduction();
				
				meanProbRep += probRep;
				
				if (probRep > maxProbRep)
					maxProbRep = probRep;
				
				if (probRep < minProbRep)
					minProbRep = probRep;

				// does this organism belong to a significant species?
				map< OrgID, SpeciesInfo, OrgID::ltOrgID >::iterator org_it = m_CurrentSpeciesMap.find( orgs[i]->getID() );
				if ((org_it != m_CurrentSpeciesMap.end()) && (org_it->second.count > SPECIES_NUMBER_SIG_THRESHOLD))
				{
					++totalNumSigOrgs;
					
					// calculate sig size stats
					meanSigSize += orgSize;
					
					if (orgSize > maxSigSize)
						maxSigSize = orgSize;
					
					if (orgSize < minSigSize)
						minSigSize = orgSize;					

					// calculate sig specificity stats
					meanSigSpecificity += spec;
					
					if (spec > maxSigSpecificity)
						maxSigSpecificity = spec;
					
					if (spec < minSigSpecificity)
						minSigSpecificity = spec;					
					
					// calculate sig age stats
					meanSigAge += age;
					
					if (age > maxSigAge)
						maxSigAge = age;
					
					if (age < minSigAge)
						minSigAge = age;
					
					// calculate sig num offspring stats
					meanSigNumOffspring += numOffspring;
				
					if (numOffspring > maxSigNumOffspring)
						maxSigNumOffspring = numOffspring;
				
					if (numOffspring < minSigNumOffspring)
						minSigNumOffspring = numOffspring;
				
					// calculate sig prob reproduction stats
					meanSigProbRep += probRep;
				
					if (probRep > maxSigProbRep)
						maxSigProbRep = probRep;

					if (probRep < minSigProbRep)
						minSigProbRep = probRep;
				}
			}
		}
	}
	
	diversity = m_CurrentSpeciesMap.size();
	
	// write to sizecounts.dat
	if (RECORD_GENOME_SIZE_FILE)
	{
		ofstream sizecounts( "sizecounts.dat", ios::app );
		sizecounts << iteration << " ";
		for (int i=0; i < GENOME_SIZE_FILE_MAX_SIZE; ++i)
			sizecounts << numgenecounts[i] << " ";
		sizecounts << endl;
		sizecounts.close();
	}
	
	// write to specificitycounts.dat
	if (RECORD_GENOME_SPECIFICITY_FILE && ORGANISM_REQUIREMENTS_EVOLVABLE)
	{
		ofstream specificitycounts( "specificitycounts.dat", ios::app );
		specificitycounts << iteration << " ";
		for (int i=0; i < LOCAL_ENVIRONMENT_VECTOR_SIZE; ++i)
			specificitycounts << speccounts[i] << " ";
		specificitycounts << endl;
		specificitycounts.close();
	}
	
	// record data for significant species (and write to speciescounts.dat if necessary)
	ofstream speciescounts;
	if (RECORD_SPECIES_COUNTS_FILE)
	{
		stringstream scname;
		getSpeciesCountsFilename( scname, m_CurrentIteration / SPECIES_COUNTS_FILE_ZIP_PERIOD );
		speciescounts.open( scname.str().c_str(), ios::app );
		speciescounts << iteration << ":";
	}
	map< OrgID, SpeciesInfo, OrgID::ltOrgID >::iterator cnt_it  = m_CurrentSpeciesMap.begin();
	map< OrgID, SpeciesInfo, OrgID::ltOrgID >::iterator cnt_end = m_CurrentSpeciesMap.end();
	while (cnt_it != cnt_end)
	{
		if (RECORD_SPECIES_COUNTS_FILE)
			speciescounts << cnt_it->first << " " << cnt_it->second.count << ";";
	
		if (cnt_it->second.count >= SPECIES_NUMBER_SIG_THRESHOLD)
			++sigDiversity;

		++cnt_it;
	}
	if (RECORD_SPECIES_COUNTS_FILE)
	{	
		speciescounts << endl;
		speciescounts.close();
	}

	// write to summary.dat
	ofstream summary( "summary.dat", ios::app );
	summary	<< iteration << " " << totalNumOrgs << " " << totalNumSigOrgs << " " << diversity	// 1-4
			<< " " << sigDiversity << " " << minSize << " " << maxSize << " "					// 5-7
			<< ((totalNumOrgs > 0) ? ((double)meanSize / (double)totalNumOrgs) : 0 )			// 8
			<< " " << minSigSize << " " << maxSigSize << " "									// 9-10
			<< ((totalNumSigOrgs > 0) ? ((double)meanSigSize / (double)totalNumSigOrgs) : 0 )	// 11
			<< " " << minProbRep << " " << maxProbRep << " "									// 12-13
			<< ((totalNumOrgs > 0) ? ((double)meanProbRep / (double)totalNumOrgs) : 0 )			// 14
			<< " " << minSigProbRep << " " << maxSigProbRep << " "								// 15-16
			<< ((totalNumSigOrgs > 0) ? ((double)meanSigProbRep / (double)totalNumSigOrgs) : 0 )// 17	
			<< " " << minAge << " " << maxAge << " "											// 18-19
			<< ((totalNumOrgs > 0) ? ((double)meanAge / (double)totalNumOrgs) : 0 )				// 20
			<< " " << minSigAge << " " << maxSigAge << " "										// 21-22
			<< ((totalNumSigOrgs > 0) ? ((double)meanSigAge / (double)totalNumSigOrgs) : 0 )	// 23
			<< " " << minNumOffspring << " " << maxNumOffspring << " "							// 24-25
			<< ((totalNumOrgs > 0) ? ((double)meanNumOffspring / (double)totalNumOrgs) : 0 )	// 26
			<< " " << minSigNumOffspring << " " << maxSigNumOffspring << " "					// 27-28
			<< ((totalNumSigOrgs > 0) ? ((double)meanSigNumOffspring / (double)totalNumSigOrgs) : 0 );// 29
	
	if (ORGANISM_REQUIREMENTS_EVOLVABLE)
	{
		summary	<< " " << minSpecificity << " " << maxSpecificity << " "						// 30-31
				<< ((totalNumOrgs > 0) ? ((double)meanSpecificity / (double)totalNumOrgs) : 0 )	// 32
				<< " " << minSigSpecificity << " " << maxSigSpecificity << " "					// 33-34
				<< ((totalNumSigOrgs > 0) ? ((double)meanSigSpecificity / (double)totalNumSigOrgs) : 0 );// 35
	}
	
	summary	<< endl;
	summary.close();

	// finally, zip species counts file if necessary
	if ((m_CurrentIteration / SPECIES_COUNTS_FILE_ZIP_PERIOD) > m_LatestSpeciesCountsIdx)
	{
		if (RECORD_SPECIES_COUNTS_FILE)
		{
			stringstream oldname;
			getSpeciesCountsFilename( oldname, m_LatestSpeciesCountsIdx );
			stringstream zipcommand;
			zipcommand << "gzip " << oldname.str() << " &";
			system( zipcommand.str().c_str() );
		}
		
		m_LatestSpeciesCountsIdx = m_CurrentIteration / SPECIES_COUNTS_FILE_ZIP_PERIOD;
	}		
}


void Simulator::getSpeciesCountsFilename( stringstream & scname, int idx )
{
	scname << "speciescounts" << setfill('0') << setw(4) << idx << ".dat";
}


void Simulator::getNewOrgID( int specificity, int numgenes, OrgID * pNewID )
{
	unsigned int tag;
	unsigned int key = getOrgIDKey( specificity, numgenes );

	hash_map<unsigned int, unsigned int>::iterator it = m_OrgIDMap.find( key );
	if (it == m_OrgIDMap.end())
	{
		tag = 0;
		m_OrgIDMap.insert(make_pair(key, 0));
	}
	else
	{
		tag = (++(it->second));
	}
	
	pNewID->m_Specificity = specificity;
	pNewID->m_NumGenes = numgenes;
	pNewID->m_Tag = tag;
}


void Simulator::saveState( int iteration )
{
	stringstream name;
	name << "niche-state-iteration-" << iteration;
	ofstream file(name.str().c_str());
	
	for (int x = 0; x < ENV_SIZE_X; ++x)
		for (int y = 0; y < ENV_SIZE_Y; ++y)
			file << getPatch(x,y) << endl;
	
	file.close();
		
	stringstream compresscommand;
	compresscommand << "gzip " << name.str() << " &";
	system( compresscommand.str().c_str() );
}


void Simulator::recordOrgBirth( OrgID & id )
{
	map< OrgID, SpeciesInfo, OrgID::ltOrgID >::iterator map_it = m_CurrentSpeciesMap.find(id);
	if (map_it == m_CurrentSpeciesMap.end())
	{
		SpeciesInfo info;
		m_CurrentSpeciesMap.insert(make_pair(id,info));
	}
	else
	{
		(map_it->second.count)++;
	}
}


void Simulator::recordOrgDeath( Organism * pOrg )
{
	map< OrgID, SpeciesInfo, OrgID::ltOrgID >::iterator map_it = m_CurrentSpeciesMap.find( pOrg->getID() );
	if (map_it == m_CurrentSpeciesMap.end())
	{
		cerr << "In Simulator::recordOrgDeath, have no record of organism " << pOrg->getID() << endl;
		exit( EXIT_FAILURE );
	}
	else
	{
		if (--(map_it->second.count) == 0)
		{
			m_CurrentSpeciesMap.erase( map_it );
		}
		else
		{
			++(map_it->second.numDead);
			(map_it->second.cumulativeAgeAtDeath) += ( 1 + m_CurrentIteration - pOrg->getTimeOfBirth() );
			(map_it->second.cumulativeNumOffspring) += pOrg->getNumOffspring();
		}
	}
}


void Simulator::loadParameterValues( const char * paramfilename )
{
	int i;
	int linenum = 1;
	bool moreToRead = true;
	bool recognised = false;
	bool comment = false;
	FILE *fp;
	char sLine[ _MaxParamFileLineLength ];
	char sErrorMsg[ _MaxParamFileLineLength ];
	char sSeparators[] = "=, \t\n";
	char *psLinePos = NULL, *psNextItem = NULL;

	_RegisterParamULong( &RNG_SEED, "RNG_SEED" );
	_RegisterParamInt( &NUM_ITERATIONS, "NUM_ITERATIONS" );
	_RegisterParamInt( &ENV_SIZE_X, "ENV_SIZE_X" );
	_RegisterParamInt( &ENV_SIZE_Y, "ENV_SIZE_Y" );
	_RegisterParamInt( &DISPERSAL_RADIUS, "DISPERSAL_RADIUS" );
	_RegisterParamInt( &NEIGHBOURHOOD_RADIUS, "NEIGHBOURHOOD_RADIUS" );
	_RegisterParamFloat( &REPRODUCTION_FUNCTION_COEFF_K, "REPRODUCTION_FUNCTION_COEFF_K" );
	_RegisterParamFloat( &REPRODUCTION_FUNCTION_COEFF_C, "REPRODUCTION_FUNCTION_COEFF_C" );
	_RegisterParamFloat( &REPRODUCTION_FUNCTION_COEFF_X, "REPRODUCTION_FUNCTION_COEFF_X" );
	_RegisterParamInt( &REPRODUCTION_FUNCTION_TYPE, "REPRODUCTION_FUNCTION_TYPE" );
	_RegisterParamInt( &DEATH_POLICY, "DEATH_POLICY" );
	_RegisterParamFloat( &DEATH_PROB_COEFF_X, "DEATH_PROB_COEFF_X" );
	_RegisterParamFloat( &LOCAL_POP_DENSITY_MAX, "LOCAL_POP_DENSITY_MAX" );
	_RegisterParamFloat( &LOCAL_POP_DENSITY_INIT, "LOCAL_POP_DENSITY_INIT" );
	_RegisterParamFloat( &MUTATION_ADD_PROB, "MUTATION_ADD_PROB" );
	_RegisterParamFloat( &MUTATION_DELETE_PROB, "MUTATION_DELETE_PROB" );
	_RegisterParamFloat( &MUTATION_CHANGE_PROB_PER_GENE, "MUTATION_CHANGE_PROB_PER_GENE" );
	_RegisterParamFloat( &MUTATION_CHANGE_PROB_REQ, "MUTATION_CHANGE_PROB_REQ" );
	_RegisterParamBool( &MUTATION_PROPORTIONAL_TO_LENGTH, "MUTATION_PROPORTIONAL_TO_LENGTH" );
	_RegisterParamBool( &ORGANISM_REQUIREMENTS_EVOLVABLE, "ORGANISM_REQUIREMENTS_EVOLVABLE" );
	_RegisterParamBool( &ENVIRONMENT_DISTURBANCE_RESTRICTED, "ENVIRONMENT_DISTURBANCE_RESTRICTED" );
	_RegisterParamInt( &ENVIRONMENT_DISTURBANCE_MAX_SPREAD, "ENVIRONMENT_DISTURBANCE_MAX_SPREAD" );
	_RegisterParamBool( &ENVIRONMENT_DISTURBANCE_ENABLED, "ENVIRONMENT_DISTURBANCE_ENABLED" );
	_RegisterParamFloat( &ORGANISM_SPECIFICITY_FRACTION_INIT, "ORGANISM_SPECIFICITY_FRACTION_INIT" );
	_RegisterParamInt( &LOCAL_ENVIRONMENT_VECTOR_SIZE, "LOCAL_ENVIRONMENT_VECTOR_SIZE" );
	_RegisterParamFloat( &LOCAL_ENVIRONMENT_VECTOR_INIT_VAL, "LOCAL_ENVIRONMENT_VECTOR_INIT_VAL" );
	_RegisterParamInt( &ABIOTIC_ENV_TYPE, "ABIOTIC_ENV_TYPE" );
	_RegisterParamInt( &NEW_GENOME_LENGTH_MIN, "NEW_GENOME_LENGTH_MIN" );
	_RegisterParamInt( &NEW_GENOME_LENGTH_MAX, "NEW_GENOME_LENGTH_MAX" );
	_RegisterParamBool( &RECORD_PHYLOGENY_FILE_FULL, "RECORD_PHYLOGENY_FILE_FULL" );
	_RegisterParamBool( &RECORD_PHYLOGENY_FILE_PARTIAL, "RECORD_PHYLOGENY_FILE_PARTIAL" );
	_RegisterParamBool( &RECORD_GENOME_BANK_FILE, "RECORD_GENOME_BANK_FILE" );
	_RegisterParamBool( &RECORD_GENOME_SIZE_FILE, "RECORD_GENOME_SIZE_FILE" );
	_RegisterParamBool( &RECORD_GENOME_SPECIFICITY_FILE, "RECORD_GENOME_SPECIFICITY_FILE" );
	_RegisterParamBool( &RECORD_SPECIES_COUNTS_FILE, "RECORD_SPECIES_COUNTS_FILE" );
	_RegisterParamBool( &RECORD_SIMULATION_STATE, "RECORD_SIMULATION_STATE" );
	_RegisterParamInt( &DATA_RECORD_PERIOD, "DATA_RECORD_PERIOD" );
	_RegisterParamInt( &SIMULATION_STATE_RECORD_PERIOD, "SIMULATION_STATE_RECORD_PERIOD" );
	_RegisterParamInt( &PHYLOGENY_PARTIAL_RECORD_PERIOD, "PHYLOGENY_PARTIAL_RECORD_PERIOD" );
	_RegisterParamInt( &PHYLOGENY_PARTIAL_RECORD_WINDOW, "PHYLOGENY_PARTIAL_RECORD_WINDOW" );
	_RegisterParamInt( &GENOME_SIZE_FILE_MAX_SIZE, "GENOME_SIZE_FILE_MAX_SIZE" );
	_RegisterParamInt( &GENOME_BANK_RECORD_THRESHOLD, "GENOME_BANK_RECORD_THRESHOLD" );
	_RegisterParamInt( &SPECIES_NUMBER_SIG_THRESHOLD, "SPECIES_NUMBER_SIG_THRESHOLD" );
	_RegisterParamInt( &SPECIES_COUNTS_FILE_ZIP_PERIOD, "SPECIES_COUNTS_FILE_ZIP_PERIOD" );
	
	// open the parameter specification file
	fp = fopen( paramfilename, "r" );
	if (fp == NULL)
	{
		cerr << "In method Simulator::loadParameterValues: Cannot open file " << paramfilename << ". Aborting." << endl;
		exit( EXIT_FAILURE );
	}

	// now go through the file a line at a time looking for parameter specifications
	while (moreToRead)
	{
		recognised = false;
		comment = false;
		
		// read next line
		moreToRead = ( fgets( sLine, _MaxParamFileLineLength, fp ) == NULL ) ? false : true;
		
		if ( moreToRead &&
		   (( strncmp( sLine, "//", 2 ) == 0 ) ||
			( sLine[0] == '\n' ) ||
			( sLine[0] == '%' ) ||
			( strspn( sLine, " \t\n" ) == strlen( sLine ) ) ) )
			comment = true;
		
		if (moreToRead && !comment)
		{
			// check for int parameters
			for (i = 0; ((i < numIntParams) && (!recognised)); i++)
			{
				if ( ( strncmp( sLine, IntParamNames[i], mymin(strlen(IntParamNames[i]), strlen(sLine)) ) ) == 0 )
				{
					psLinePos = sLine + strlen( IntParamNames[i] );
					sprintf( sErrorMsg, "Problem with specification for parameter %s on line %d of file %s", IntParamNames[i], linenum, paramfilename );
					psNextItem = gstrtokfromfile( psLinePos, sSeparators, paramfilename, fp, sErrorMsg );
					*(IntParams[i]) = atoi( psNextItem );
					recognised = true;
				}
			}
			
			// check for unsigned int parameters
			for (i = 0; ((i < numUIntParams) && (!recognised)); i++)
			{
				if ( ( strncmp( sLine, UIntParamNames[i], mymin(strlen(UIntParamNames[i]), strlen(sLine)) ) ) == 0 )
				{
					psLinePos = sLine + strlen( UIntParamNames[i] );
					sprintf( sErrorMsg, "Problem with specification for parameter %s on line %d of file %s", UIntParamNames[i], linenum, paramfilename );
					psNextItem = gstrtokfromfile( psLinePos, sSeparators, paramfilename, fp, sErrorMsg );
					*(UIntParams[i]) = (unsigned int) atoi( psNextItem );
					recognised = true;
				}
			}

			// check for unsigned long parameters
			for (i = 0; ((i < numULongParams) && (!recognised)); i++)
			{
				if ( ( strncmp( sLine, ULongParamNames[i], mymin(strlen(ULongParamNames[i]), strlen(sLine)) ) ) == 0 )
				{
					psLinePos = sLine + strlen( ULongParamNames[i] );
					sprintf( sErrorMsg, "Problem with specification for parameter %s on line %d of file %s", IntParamNames[i], linenum, paramfilename );
					psNextItem = gstrtokfromfile( psLinePos, sSeparators, paramfilename, fp, sErrorMsg );
					*(ULongParams[i]) = atol( psNextItem );
					recognised = true;
				}
			}			
			
			// check for float parameters
			for (i = 0; ((i < numFloatParams) && (!recognised)); i++)
			{
				if ( ( strncmp( sLine, FloatParamNames[i], mymin(strlen(FloatParamNames[i]), strlen(sLine)) ) ) == 0 )
				{
					psLinePos = sLine + strlen( FloatParamNames[i] );
					sprintf( sErrorMsg, "Problem with specification for parameter %s on line %d of file %s", FloatParamNames[i], linenum, paramfilename );
					psNextItem = gstrtokfromfile( psLinePos, sSeparators, paramfilename, fp, sErrorMsg );
					*(FloatParams[i]) = atof( psNextItem );
					recognised = true;
				}
			}
			
			// check for bool parameters
			for (i = 0; ((i < numBoolParams) && (!recognised)); i++)
			{
				if ( ( strncmp( sLine, BoolParamNames[i], mymin(strlen(BoolParamNames[i]), strlen(sLine)) ) ) == 0 )
				{
					psLinePos = sLine + strlen( BoolParamNames[i] );
					sprintf( sErrorMsg, "Problem with specification for parameter %s on line %d of file %s", BoolParamNames[i], linenum, paramfilename );
					psNextItem = gstrtokfromfile( psLinePos, sSeparators, paramfilename, fp, sErrorMsg );
					if (strcmp( psNextItem, "true" ) == 0)
					{
						*(BoolParams[i]) = true;
						recognised = true;
					}
					else if (strcmp( psNextItem, "false" ) == 0)
					{
						*(BoolParams[i]) = false;
						recognised = true;
					}
				}
			}
			
			// check for string parameters
			for (i = 0; ((i < numStringParams) && (!recognised)); i++)
			{
				if ( ( strncmp( sLine, StringParamNames[i], mymin(strlen(StringParamNames[i]), strlen(sLine)) ) ) == 0 )
				{
					psLinePos = sLine + strlen( StringParamNames[i] );
					sprintf( sErrorMsg, "Problem with specification for parameter %s on line %d of file %s", StringParamNames[i], linenum, paramfilename );
					psNextItem = gstrtokfromfile( psLinePos, sSeparators, paramfilename, fp, sErrorMsg );
					strcpy( StringParams[i], psNextItem );
					recognised = true;
				}
			}

			if (!recognised)
			{
				sprintf( sErrorMsg, "Warning! Unable to interpret line %d of parameter file %s. Ignoring line.\n", linenum, paramfilename );
				fprintf( stderr, sErrorMsg );
			}
		}

		linenum++;
	}

	fclose( fp );
	
	// Some variables have values calculated from other variables. These may need to be adjusted now.
	setDerivedVariables();
}


void Simulator::recordParameterValues( ostream& os )
{
	int i;
	for (i=0; i<numIntParams; ++i)
		os << IntParamNames[i] << "=" << *(IntParams[i]) << endl;
	for (i=0; i<numUIntParams; ++i)
		os << UIntParamNames[i] << "=" << *(UIntParams[i]) << endl;
	for (i=0; i<numULongParams; ++i)
		os << ULongParamNames[i] << "=" << *(ULongParams[i]) << endl;	
	for (i=0; i<numFloatParams; ++i)
		os << FloatParamNames[i] << "=" << *(FloatParams[i]) << endl;
	for (i=0; i<numBoolParams; ++i)
		os << BoolParamNames[i] << "=" << ((*(BoolParams[i])) ? "true" : "false") << endl;
	for (i=0; i<numStringParams; ++i)
		os << StringParamNames[i] << "=" << (StringParams[i]) << endl;	
}


char * Simulator::gstrtokfromfile( char * line, char * separators, const char * filename, FILE* fp, char * errormsg )
{
	// perform a graceful strtok on a line from an input file
	
	char *pResult;
	char msg[ _MaxParamFileLineLength ];
	
	if ( (pResult = strtok( line, separators )) == NULL )
	{
		fclose( fp );
		sprintf( msg, "Error reading file %s: %s. Aborting.", filename, errormsg );
		cerr << msg << endl;
		exit( EXIT_FAILURE );
	}
	return  pResult;
}


void Simulator::setDerivedVariables()
{
	NEIGHBOURHOOD_SIZE = (NEIGHBOURHOOD_RADIUS * 2 + 1) * (NEIGHBOURHOOD_RADIUS * 2 + 1);
	if (ENVIRONMENT_DISTURBANCE_MAX_SPREAD > Gene::MAX_SPREAD)
		ENVIRONMENT_DISTURBANCE_MAX_SPREAD = Gene::MAX_SPREAD;
}


bool Simulator::inPhylogenyRecordWindow()
{
	static bool previousInWindow = true;
	static int fileIdx = 0;
	
	bool inWindow = false;
	
	if (RECORD_PHYLOGENY_FILE_FULL)
	{
		inWindow = true;
	}
	else if (RECORD_PHYLOGENY_FILE_PARTIAL)
	{
		inWindow = ((m_CurrentIteration % PHYLOGENY_PARTIAL_RECORD_PERIOD) >= (PHYLOGENY_PARTIAL_RECORD_PERIOD - PHYLOGENY_PARTIAL_RECORD_WINDOW));
		
		if ((!inWindow) && (previousInWindow) && (fileIdx > 0))
		{
			stringstream command1, command2;
			command1 << "mv phylogeny.dat phylogeny" << setfill('0') << setw(4) << fileIdx << ".dat";
			system( command1.str().c_str() );
			command2 << "gzip phylogeny" << setfill('0') << setw(4) << fileIdx << ".dat &";
			system( command2.str().c_str() );
		}
		
		if (inWindow && !previousInWindow)
		{
			++fileIdx;
			//ofstream phyl( "phylogeny.dat" ); // open file without ios::app to erase its contents
			//phyl.close();
		}
	}
	
	previousInWindow = inWindow;
	
	return inWindow;
}
