// ande.cc
// Process data from speciescounts.dat and phylogeny.dat to generate ancestor-descendant test data
//
// Usage: ande g t 
// where g = generation number
//       t = threshold conc above which species are included in test
//       s = min size to consider
//
// compile with: gcc -Wall -g ande.cc OrgID.cc -I../include -lstdc++ -o ../bin/ande
//

#include "OrgID.h"
#include <fstream>
#include <set>
#include <string>
#include <sstream>
#include <cstdlib>
#include <cassert>


int main( int argc, const char *argv[] )
{
	if (argc != 4)
	{
		cerr << "ANDE usage: ande g t s" << endl
		<< "where\tg = generation number" << endl
		<< "\tt = threshold conc above which species are included in test" << endl
		<< "\ts = min size to consider" << endl;
		exit( EXIT_FAILURE );
	}
	
	int g = atoi(argv[1]);
	int t = atoi(argv[2]);
	int s = atoi(argv[3]);	
	
	ifstream concfile( "speciescounts.dat" );
	
	if (concfile.bad())
	{
		cerr << "Could not open speciescounts.dat file!" << endl;
		exit( EXIT_FAILURE );
	}
	
	// go through speciescounts.dat
	// record all species with counts above threshold t into sigspecies set
	
	set< OrgID, OrgID::ltOrgID > sigspecies;
	string lineStr;
	OrgID id;
	int linenum = 1;
	int gen, count;
	bool lineFound = false;
	
	while ( getline(concfile, lineStr) && !lineFound )  // this version of getline defined in <string>
	{
		stringstream lineStrStr( lineStr );
		lineStrStr >> gen;
		if (!lineStrStr)
		{
			cout << "Error reading file speciescounts.dat, line " << linenum << endl;
			exit( EXIT_FAILURE );
		}
		
		if (gen == g)
		{
			cout << "Generation: " << gen << endl;
			while (! lineStrStr.eof() )
			{
				lineStrStr >> id >> count;
				if (count >= t)
					sigspecies.insert(id);
			}
			lineFound = true;
		}

		++linenum;
	}
	
	concfile.close();

	if (!lineFound)
	{
		cerr << "Unable to find data for generation " << g << " in speciescounts.dat!" << endl;
		exit( EXIT_FAILURE );
	}
	
	// now, for each species in sigspecies list, look for its ancestor in phylogeny.dat
	ifstream phylfile( "phylogeny.dat" );
	
	if (phylfile.bad())
	{
		cerr << "Could not open speciescounts.dat file!" << endl;
		exit( EXIT_FAILURE );
	}
	
	OrgID parentID;
	
	int numDecs = 0;
	int numSame = 0;
	int numIncs = 0;
	
	string idStr;
	
	cout << "Number of significant species=" << sigspecies.size() << endl;
	
	while ( getline(phylfile, lineStr) )  // this version of getline defined in <string>
	{
		int commapos = lineStr.find( ",", 0 );
		if (commapos > 0 )
		{
			idStr = lineStr.substr( 0, commapos );
			id.set( idStr );
			idStr = lineStr.substr( commapos+1, lineStr.size() - commapos - 1 );
			parentID.set( idStr );
			//cout << lineStr.c_str() << ", species=" << id << ", parent=" << parentID << endl;
			set< OrgID, OrgID::ltOrgID >::iterator org_it = sigspecies.find(id);
			if (org_it != sigspecies.end())
			{
				cout << "Found a sig species, " << id << ", numGenes=" << id.m_NumGenes << endl;
				if (id.m_NumGenes > s)
				{
					if (id.m_NumGenes > parentID.m_NumGenes)
						++numIncs;
					else if (id.m_NumGenes == parentID.m_NumGenes)
						++numSame;
					else
						++numDecs;
				}
			}
		}
	}
	
	cout << "Number of decreases=" << numDecs << ", Number with no change=" << numSame
		<< ", Number of increases=" << numIncs << endl;

	ofstream ans( "ande.dat", ios::app );
	ans << g << " " << t << " " << s << " " << numDecs << " " << numSame << " " << numIncs << endl;
	ans.close();

	return EXIT_SUCCESS;
}
